// Roger P Wright  - preprocessor for a Wylie to Unicode parser and convertor.
// Copyright (C) 2003 Roger P Wright
//
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  Contact the author at Roger_Wright@attglobal.net
//
//  See the gnu_lic.txt file for the complete license.
//
#include <stdio.h>
#include <string.h>
#include "strcmpl.c"
FILE *ifp, *ofp ;
char iline [4096] ;
char oline [4096] ;
char syll[128] ;
char wylie_syll[130] ;
char slot[13] ;
char vowelslot[64] ;
#define NUMPREFIX 367
char *input_prefix_cluster_table[NUMPREFIX] = {
"k",
"ky",
"kr",
"kl",
"kw",
"dk",
"dky",
"dkr",
"bk",
"bky",
"bkr",
"bkl",
"\'k",
"rk",
"rky",
"lk",
"sk",
"sky",
"skr",
"brk",
"brky",
"bsk",
"bsky",
"bskr",
"K",
"Ky",
"Kr",
"Kw",
"mK",
"mKy",
"mKr",
"\'K",
"\'Ky",
"\'Kr",
"rK",     /*skt trans*/
"rKw",    /*skt trans*/
"g",
"gy",
"gr",
"grw",
"gl",
"gw",
"dg",
"dgy",
"dgr",
"bg",
"bgy",
"bgr",
"mg",
"mgy",
"mgr",
"\'g",
"\'gy",
"\'gr",
"rg",
"rgy",
"rgr",    /*skt trans*/
"rgw",
"lg",
"sg",
"sgy",
"sgr",
"brg",
"brgy",
"bsg",
"bsgy",
"bsgr",
"G",      /*skt trans*/
"Gy",     /*skt trans*/
"Gr",     /*skt trans*/
"Gw",     /*skt trans*/
"rG",     /*skt trans*/
"rGy",    /*skt trans*/
"R",
"Ry",     /*skt trans*/
"Rr",     /*skt trans*/
"Rw",     /*skt trans*/
"dR",
"mR",
"rR",
"lR",
"sR",
"brR",
"bsR",
"c",
"cy",     /*skt trans*/
"cr",     /*skt trans*/
"cw",
"gc",
"bc",
"lc",
"C",
"Cy",     /*skt trans*/
"Cr",     /*skt trans*/
"Cw",     /*skt trans*/
"mC",
"\'C",
"j",
"jy",     /*skt trans*/
"jr",     /*skt trans*/
"jw",     /*skt trans*/
"bj",
"mj",
"\'j",
"rj",
"lj",
"brj",
"N",
"Ny",     /*skt trans*/
"Nr",     /*skt trans*/
"Nw",     /*skt trans*/
"gN",
"mN",
"rN",
"sN",
"brN",
"bsN",
"V",   /* .t */
"Vy",     /*skt trans*/
"Vr",     /*skt trans*/
"Vrw",    /*skt trans*/
"Vw",     /*skt trans*/
"rV",     /*skt trans*/
"rVy",    /*skt trans*/
"W",   /*.th */
"Wy",     /*skt trans*/
"Wr",     /*skt trans*/
"Ww",     /*skt trans*/
"rW",     /*skt trans*/
"X",   /* .d */
"Xy",     /*skt trans*/
"Xr",     /*skt trans*/
"Xw",     /*skt trans*/
"rX",     /*skt trans*/
"Y",   /*.dh */
"Yy",     /*skt trans*/
"Yr",     /*skt trans*/
"Yw",     /*skt trans*/
"rY",     /*skt trans*/
"rYy",    /*skt trans*/
"L",   /* .n */
"Ly",     /*skt trans*/
"Lr",     /*skt trans*/
"Lw",     /*skt trans*/
"rL",     /*skt trans*/
"rLy",    /*skt trans*/
"t",
"ty",     /*skt trans*/
"tr",
"trw",    /*skt trans*/
"tw",
"gt",
"bt",
"rt",
"rty",    /*skt trans*/
"rtr",    /*skt trans*/
"rtw",    /*skt trans*/
"lt",
"st",
"brt",
"blt",
"bst",
"T",
"Ty",     /*skt trans*/
"Tr",
"Tw",     /*skt trans*/
"mT",
"\'T",
"rT",     /*skt trans*/
"rTy",    /*skt trans*/
"d",
"dy",     /*skt trans*/
"dr",
"drw",
"dw",
"gd",
"bd",
"md",
"\'d",
"\'dr",
"rd",
"rdr",    /*skt trans*/
"ld",
"sd",
"brd",
"bld",
"bsd",
"D",
"Dy",     /*skt trans*/
"Dr",     /*skt trans*/
"Dw",     /*skt trans*/
"rD",     /*skt trans*/
"rDy",    /*skt trans*/
"rDr",    /*skt trans*/
"rDw",    /*skt trans*/
"n",
"ny",     /*skt trans*/
"nyw",     /*skt trans*/
"nr",
"nw",     /*skt trans*/
"gn",
"bn",
"mn",
"rn",
"sn",
"snr",
"brn",
"bsn",
"p",
"py",
"pr",
"pw",     /*skt trans*/
"dp",
"dpy",
"dpr",
"rp",
"rpy",    /*skt trans*/
"lp",
"sp",
"spy",
"spr",
"P",
"Py",
"Pyw",
"Pr",
"Pw",     /*skt trans*/
"\'P",
"\'Py",
"\'Pr",
"rP",     /*skt trans*/
"rPy",    /*skt trans*/
"b",
"by",
"br",
"bl",
"bw",     /*skt trans*/
"db",
"dby",
"dbr",
"\'b",
"\'by",
"\'br",
"rb",
"rby",    /*skt trans*/
"rbw",    /*skt trans*/
"lb",
"sb",
"sby",
"sbr",
"B",
"By",     /*skt trans*/
"Br",     /*skt trans*/
"Bw",     /*skt trans*/
"rB",     /*skt trans*/
"m",
"my",
"mr",
"mw",     /*skt trans*/
"dm",
"dmy",
"rm",
"rmy",
"sm",
"smy",
"smr",
"q",      /* ts */
"qy",     /*skt trans*/
"qr",     /*skt trans*/
"qw",
"gq",
"bq",
"rq",
"rqy",    /*skt trans*/
"rqw",    /*skt trans*/
"sq",
"brq",
"bsq",
"Q",      /* tsh */
"Qy",     /*skt trans*/
"Qr",     /*skt trans*/
"Qw",
"mQ",
"\'Q",
"rQ",     /*skt trans*/
"x",      /* dz */
"xy",     /*skt trans*/
"xr",     /*skt trans*/
"xw",     /*skt trans*/
"bx",
"mx",
"\'x",
"rx",
"rxy",    /*skt trans*/
"brx",
"J",      /* dzh */
"Jy",     /*skt trans*/
"Jr",     /*skt trans*/
"Jw",     /*skt trans*/
"rJ",     /*skt trans*/
"w",
"wy",     /*skt trans*/
"wr",     /*skt trans*/
"ww",     /*skt trans*/
"Z",      /* zh */
"Zw",
"gZ",
"bZ",
"z",
"zl",
"zw",
"gz",
"bz",
"bzl",
"\'",
"y",
"yy",     /*skt trans*/
"yr",     /*skt trans*/
"yw",     /*skt trans*/
"ryy",    /*skt trans*/
"g.y",
"r",
"ry",     /*skt trans*/
"rr",     /*skt trans*/
"rl",
"rw",
"brl",
"l",
"ly",     /*skt trans*/
"lr",     /*skt trans*/
"lw",
"f",      /* sh */
"fy",     /*skt trans*/
"fr",
"fl",     /*skt trans*/
"fw",
"gf",
"bf",
"rf",     /*skt trans*/
"rfy",    /*skt trans*/
"rfw",    /*skt trans*/
"S",      /* .s */
"Sy",     /*skt trans*/
"Sr",     /*skt trans*/
"Sw",     /*skt trans*/
"rS",     /*skt trans*/
"rSy",    /*skt trans*/
"F",      /* k.s */
"Fy",     /*skt trans*/
"Fr",     /*skt trans*/
"Fw",     /*skt trans*/
"s",
"sy",     /*skt trans*/
"sr",
"sl",
"sw",
"gs",
"bs",
"bsr",
"bsl",
"rs",     /*skt trans*/
"h",
"hy",     /*skt trans*/
"hr",
"hw",
"rh",     /*skt trans*/
"rhr",    /*skt trans*/
"lh"
} ;
char *output_prefix_cluster_table[NUMPREFIX] = {
"  k  ",
"  ky ",
"  kr ",
"  kl ",
"  kw ",
"d k  ",
"d ky ",
"d kr ",
"b k  ",
"b ky ",
"b kr ",
"b kl ",
"\' k  ",
" rk  ",
" rky ",
" lk  ",
" sk  ",
" sky ",
" skr ",
"brk  ",
"brky ",
"bsk  ",
"bsky ",
"bskr ",
"  K  ",
"  Ky ",
"  Kr ",
"  Kw ",
"m K  ",
"m Ky ",
"m Kr ",
"\' K  ",
"\' Ky ",
"\' Kr ",
" rK  ",  /*skt trans*/
" rKw ",  /*skt trans*/
"  g  ",
"  gy ",
"  gr ",
"  grw",
"  gl ",
"  gw ",
"d g  ",
"d gy ",
"d gr ",
"b g  ",
"b gy ",
"b gr ",
"m g  ",
"m gy ",
"m gr ",
"\' g  ",
"\' gy ",
"\' gr ",
" rg  ",
" rgr ",  /*skt trans*/
" rgy ",
" rgw ",
" lg  ",
" sg  ",
" sgy ",
" sgr ",
"brg  ",
"brgy ",
"bsg  ",
"bsgy ",
"bsgr ",
"  G  ",  /*skt trans*/
"  Gy ",  /*skt trans*/
"  Gr ",  /*skt trans*/
"  Gw ",  /*skt trans*/
" rG  ",  /*skt trans*/
" rGy ",  /*skt trans*/
"  R  ",
"  Ry ",  /*skt trans*/
"  Rr ",  /*skt trans*/
"  Rw ",  /*skt trans*/
"d R  ",
"m R  ",
" rR  ",
" lR  ",
" sR  ",
"brR  ",
"bsR  ",
"  c  ",
"  cy ",  /*skt trans*/
"  cr ",  /*skt trans*/
"  cw ",
"g c  ",
"b c  ",
" lc  ",
"  C  ",
"  Cy ",  /*skt trans*/
"  Cr ",  /*skt trans*/
"  Cw ",  /*skt trans*/
"m C  ",
"\' C  ",
"  j  ",
"  jy ",  /*skt trans*/
"  jr ",  /*skt trans*/
"  jw ",  /*skt trans*/
"b j  ",
"m j  ",
"\' j  ",
" rj  ",
" lj  ",
"brj  ",
"  N  ",
"  Ny ",  /*skt trans*/
"  Nr ",  /*skt trans*/
"  Nw ",  /*skt trans*/
"g N  ",
"m N  ",
" rN  ",
" sN  ",
"brN  ",
"bsN  ",
"  V  ",  /*  .t */
"  Vy ",  /*skt trans*/
"  Vr ",  /*skt trans*/
"  Vrw",  /*skt trans*/
"  Vw ",  /*skt trans*/
" rV  ",  /*skt trans*/
" rVy ",  /*skt trans*/
"  W  ",  /* .th */
"  Wy ",  /*skt trans*/
"  Wr ",  /*skt trans*/
"  Ww ",  /*skt trans*/
" rW  ",  /*skt trans*/
"  X  ",  /*  .d */
"  Xy ",  /*skt trans*/
"  Xr ",  /*skt trans*/
"  Xw ",  /*skt trans*/
" rX  ",  /*skt trans*/
"  Y  ",  /* .dh */
"  Yy ",  /*skt trans*/
"  Yr ",  /*skt trans*/
"  Yw ",  /*skt trans*/
" rY  ",  /*skt trans*/
" rYy ",  /*skt trans*/
"  L  ",  /*  .n */
"  Ly ",  /*skt trans*/
"  Lr ",  /*skt trans*/
"  Lw ",  /*skt trans*/
" rL  ",  /*skt trans*/
" rLy ",  /*skt trans*/
"  t  ",
"  ty ",  /*skt trans*/
"  tr ",
"  trw ",  /*skt trans*/
"  tw ",
"g t  ",
"b t  ",
" rt  ",
" rty ",  /*skt trans*/
" rtr ",  /*skt trans*/
" rtw ",  /*skt trans*/
" lt  ",
" st  ",
"brt  ",
"blt  ",
"bst  ",
"  T  ",
"  Ty ",  /*skt trans*/
"  Tr ",
"  Tw ",  /*skt trans*/
"m T  ",
"\' T  ",
" rT  ",  /*skt trans*/
" rTy ",  /*skt trans*/
"  d  ",
"  dy ",  /*skt trans*/
"  dr ",
"  drw",
"  dw ",
"g d  ",
"b d  ",
"m d  ",
"\' d  ",
"\' dr ",
" rd  ",
" rdr ",  /*skt trans*/
" ld  ",
" sd  ",
"brd  ",
"bld  ",
"bsd  ",
"  D  ",
"  Dy ",  /*skt trans*/
"  Dr ",  /*skt trans*/
"  Dw ",  /*skt trans*/
" rD  ",  /*skt trans*/
" rDy ",  /*skt trans*/
" rDr ",  /*skt trans*/
" rDw ",  /*skt trans*/
"  n  ",
"  ny ",  /*skt trans*/
"  nyw",  /*skt trans*/
"  nr ",
"  nw ",  /*skt trans*/
"g n  ",
"b n  ",
"m n  ",
" rn  ",
" sn  ",
" snr ",
"brn  ",
"bsn  ",
"  p  ",
"  py ",
"  pr ",
"  pw ",  /*skt trans*/
"d p  ",
"d py ",
"d pr ",
" rp  ",
" rpy ",  /*skt trans*/
" lp  ",
" sp  ",
" spy ",
" spr ",
"  P  ",
"  Py ",
"  Pyw",
"  Pr ",
"  Pw ",  /*skt trans*/
"\' P  ",
"\' Py ",
"\' Pr ",
" rP  ",  /*skt trans*/
" rPy ",  /*skt trans*/
"  b  ",
"  by ",
"  br ",
"  bl ",
"  bw ",  /*skt trans*/
"d b  ",
"d by ",
"d br ",
"\' b  ",
"\' by ",
"\' br ",
" rb  ",
" rby ",  /*skt trans*/
" rbw ",  /*skt trans*/
" lb  ",
" sb  ",
" sby ",
" sbr ",
"  B  ",
"  By ",  /*skt trans*/
"  Br ",  /*skt trans*/
"  Bw ",  /*skt trans*/
" rB  ",
"  m  ",
"  my ",
"  mr ",
"  mw ",  /*skt trans*/
"d m  ",
"d my ",
" rm  ",
" rmy ",
" sm  ",
" smy ",
" smr ",
"  q  ",
"  qy ",  /*skt trans*/
"  qr ",  /*skt trans*/
"  qw ",
"g q  ",
"b q  ",
" rq  ",
" rqy ",  /*skt trans*/
" rqw ",  /*skt trans*/
" sq  ",
"brq  ",
"bsq  ",
"  Q  ",
"  Qy ",  /*skt trans*/
"  Qr ",  /*skt trans*/
"  Qw ",
"m Q  ",
"\' Q  ",
" rQ  ",  /*skt trans*/
"  x  ",
"  xy ",  /*skt trans*/
"  xr ",  /*skt trans*/
"  xw ",  /*skt trans*/
"b x  ",
"m x  ",
"\' x  ",
" rx  ",
" rxy ",  /*skt trans*/
"brx  ",
"  J  ",
"  Jy ",  /*skt trans*/
"  Jr ",  /*skt trans*/
"  Jw ",  /*skt trans*/
" rJ  ",  /*skt trans*/
"  w  ",
"  wy ",  /*skt trans*/
"  wr ",  /*skt trans*/
"  ww ",  /*skt trans*/
"  Z  ",
"  Zw ",
"g Z  ",
"b Z  ",
"  z  ",
"  zl ",
"  zw ",
"g z  ",
"b z  ",
"b zl ",
"  \'  ",
"  y  ",
"  yy ",  /*skt trans*/
"  yr ",  /*skt trans*/
"  yw ",  /*skt trans*/
" ryy ",  /*skt trans*/
"g.y  ",
"  r  ",
"  ry ",  /*skt trans*/
"  rr ",  /*skt trans*/
"  rl ",
"  rw ",
"b rl ",
"  l  ",
"  ly ",  /*skt trans*/
"  lr ",  /*skt trans*/
"  lw ",
"  f  ",
"  fy ",  /*skt trans*/
"  fr ",
"  fl ",  /*skt trans*/
"  fw ",
"g f  ",
"b f  ",
" rf  ",  /*skt trans*/
" rfy ",  /*skt trans*/
" rfw ",  /*skt trans*/
"  S  ",
"  Sy ",  /*skt trans*/
"  Sr ",  /*skt trans*/
"  Sw ",  /*skt trans*/
" rS  ",  /*skt trans*/
" rSy ",  /*skt trans*/
"  F  ",
"  Fy ",  /*skt trans*/
"  Fr ",  /*skt trans*/
"  Fw ",  /*skt trans*/
"  s  ",
"  sy ",  /*skt trans*/
"  sr ",
"  sl ",
"  sw ",
"g s  ",
"b s  ",
"b sr ",
"b sl ",
" rs  ",  /*skt trans*/
"  h  ",
"  hy ",  /*skt trans*/
"  hr ",
"  hw ",
" rh  ",  /*skt trans*/
" rhr ",  /*skt trans*/
" lh  "
} ;
char digits[13] = "0123456789()" ;
char normal_mode_front_vowels[15] = "aAiIuUeEoO\x1b\x11\x18\x19" ;
                  /* .r,.R,.l,.L */
char normal_mode_vowels[16] = "aAiIuUeEoO\x1b\x11\x18\x19+" ;
                  /* .r,.R,.l,.L and + for stacking */
char normal_mode_suffix[14] = "gRdnbm\'rlsMH." ;
         /* g ng, d, n, b, m, ', r, l, s, anusvara, visaraga, period */
char normal_mode_postsuffix[3] = "ds" ; /* d, s */
char multi_vowel_mode_front_vowels[15] = "aAiIuUeEoO\x1b\x11\x18\x19" ;
                  /* .r,.R,.l,.L */
char multi_vowel_mode_vowels[16] = "aAiIuUeEoO\x1b\x11\x18\x19+" ;
                  /* .r,.R,.l,.L,+ */
char multi_vowel_mode_suffix[5] = "RmMH" ;
         /* only Visaraga and anusvara allowed for sanskrit and ng and m for tibetan */
char multi_vowel_mode_postsuffix[1] = "" ; /* none allowed in multi-vowel mode */
/* set mode to normal mode vowel processing */
char *front_vowels = &(normal_mode_front_vowels[0]) ;
char *vowels = &(normal_mode_vowels[0]) ;
char *suffix = &(normal_mode_suffix[0]) ;
char *postsuffix = &(normal_mode_postsuffix[0]) ;

char subfixed[5] = "yrlw" ;  /* y, r, l, w */
char subsubfixed[2] = "w" ;  /* w only*/
char superfixed[5] = "rls." ;  /* r, l, s, . (for g.y) */
char root_in[44] = "kKgGRcCjNVWXYLtTdDnpPbBmqQxJwZz\'yrlfSshaFMH" ;
                   /*k kh g gh ng c ch j ny .t .th .d .dh
                     .n t th d dh n p ph b bh m ts tsh dz dzh w zh z ' y
                     r l sh .s s h front-A k.s anusvara visaraga */
char upper_root_out[43][7] = {
#define upper_root_k  "\x40\x0f"
                      upper_root_k,
#define upper_root_kh "\x41\x0f"
                      upper_root_kh,
#define upper_root_g  "\x42\x0f"
                      upper_root_g,
#define upper_root_gh "\x43\x0f"
                      upper_root_gh,
#define upper_root_ng "\x44\x0f"
                      upper_root_ng,
#define upper_root_c  "\x45\x0f"
                      upper_root_c,
#define upper_root_ch "\x46\x0f"
                      upper_root_ch,
#define upper_root_j  "\x47\x0f"
                      upper_root_j,
#define upper_root_ny "\x49\x0f"
                      upper_root_ny,
#define upper_root_T  "\x4a\x0f"
                      upper_root_T,
#define upper_root_TH "\x4b\x0f"
                      upper_root_TH,
#define upper_root_D  "\x4c\x0f"
                      upper_root_D,
#define upper_root_DH "\x4d\x0f"
                      upper_root_DH,
#define upper_root_N  "\x4e\x0f"
                      upper_root_N,
#define upper_root_t  "\x4f\x0f"
                      upper_root_t,
#define upper_root_th "\x50\x0f"
                      upper_root_th,
#define upper_root_d  "\x51\x0f"
                      upper_root_d,
#define upper_root_dh "\x52\x0f"
                      upper_root_dh,
#define upper_root_n  "\x53\x0f"
                      upper_root_n,
#define upper_root_p  "\x54\x0f"
                      upper_root_p,
#define upper_root_ph "\x55\x0f"
                      upper_root_ph,
#define upper_root_b  "\x56\x0f"
                      upper_root_b,
#define upper_root_bh "\x57\x0f"
                      upper_root_bh,
#define upper_root_m  "\x58\x0f"
                      upper_root_m,
#define upper_root_ts "\x59\x0f"
                      upper_root_ts,
#define upper_root_tsh "\x5a\x0f"
                      upper_root_tsh,
#define upper_root_dz "\x5b\x0f"
                      upper_root_dz,
#define upper_root_dzh "\x5c\x0f"
                      upper_root_dzh,
#define upper_root_w  "\x5d\x0f"
                      upper_root_w,
#define upper_root_zh "\x5e\x0f"
                      upper_root_zh,
#define upper_root_z  "\x5f\x0f"
                      upper_root_z,
#define upper_root_achung  "\x60\x0f"
                      upper_root_achung,
#define upper_root_y  "\x61\x0f"
                      upper_root_y,
#define upper_root_r  "\x62\x0f"
                      upper_root_r,
#define upper_root_l  "\x63\x0f"
                      upper_root_l,
#define upper_root_sh "\x64\x0f"
                      upper_root_sh,
#define upper_root_S  "\x65\x0f"
                      upper_root_S,
#define upper_root_s  "\x66\x0f"
                      upper_root_s,
#define upper_root_h  "\x67\x0f"
                      upper_root_h,
#define upper_root_a  "\x68\x0f"
                      upper_root_a,
#define upper_root_kS "\x69\x0f"
                      upper_root_kS,
#define upper_root_anusvara  "\x7e\x0f"
                      upper_root_anusvara,
#define upper_root_visaraga  "\x7f\x0f"
                      upper_root_visaraga,
                       } ;
char lower_root_out[43][7] = {
#define lower_root_k  "\x90\x0f"
                      lower_root_k,
#define lower_root_kh "\x91\x0f"
                      lower_root_kh,
#define lower_root_g  "\x92\x0f"
                      lower_root_g,
#define lower_root_gh "\x93\x0f"
                      lower_root_gh,
#define lower_root_ng "\x94\x0f"
                      lower_root_ng,
#define lower_root_c  "\x95\x0f"
                      lower_root_c,
#define lower_root_ch "\x96\x0f"
                      lower_root_ch,
#define lower_root_j  "\x97\x0f"
                      lower_root_j,
#define lower_root_ny "\x99\x0f"
                      lower_root_ny,
#define lower_root_T  "\x9a\x0f"
                      lower_root_T,
#define lower_root_TH "\x9b\x0f"
                      lower_root_TH,
#define lower_root_D  "\x9c\x0f"
                      lower_root_D,
#define lower_root_DH "\x9d\x0f"
                      lower_root_DH,
#define lower_root_N  "\x9e\x0f"
                      lower_root_N,
#define lower_root_t  "\x9f\x0f"
                      lower_root_t,
#define lower_root_th "\xa0\x0f"
                      lower_root_th,
#define lower_root_d  "\xa1\x0f"
                      lower_root_d,
#define lower_root_dh "\xa2\x0f"
                      lower_root_dh,
#define lower_root_n  "\xa3\x0f"
                      lower_root_n,
#define lower_root_p  "\xa4\x0f"
                      lower_root_p,
#define lower_root_ph "\xa5\x0f"
                      lower_root_ph,
#define lower_root_b  "\xa6\x0f"
                      lower_root_b,
#define lower_root_bh "\xa7\x0f"
                      lower_root_bh,
#define lower_root_m  "\xa8\x0f"
                      lower_root_m,
#define lower_root_ts "\xa9\x0f"
                      lower_root_ts,
#define lower_root_tsh "\xaa\x0f"
                      lower_root_tsh,
#define lower_root_dz "\xab\x0f"
                      lower_root_dz,
#define lower_root_dzh "\xac\x0f"
                      lower_root_dzh,
#define lower_root_w  "\xba\x0f"
                      lower_root_w,
#define lower_root_zh "\xae\x0f"
                      lower_root_zh,
#define lower_root_z  "\xaf\x0f"
                      lower_root_z,
#define lower_root_achung  "\xb0\x0f"
                      lower_root_achung,
#define lower_root_y  "\xbb\x0f"
                      lower_root_y,
#define lower_root_r  "\xbc\x0f"
                      lower_root_r,
#define lower_root_l  "\xb3\x0f"
                      lower_root_l,
#define lower_root_sh "\xb4\x0f"
                      lower_root_sh,
#define lower_root_S  "\xb5\x0f"
                      lower_root_S,
#define lower_root_s  "\xb6\x0f"
                      lower_root_s,
#define lower_root_h  "\xb7\x0f"
                      lower_root_h,
#define lower_root_a  "\xb8\x0f"
                      lower_root_a,
#define lower_root_kS "\xb9\x0f"
                      lower_root_kS,
#define lower_root_anusvara  "\x7e\x0f"
                      lower_root_anusvara,
#define lower_root_visaraga  "\x7f\x0f"
                      lower_root_visaraga,
                       } ;
char prefix[6] = "gdbm\'" ;  /* g, d, b, m, & "'" */
char wylie_encoded[35] =
                         "KGRCvNVWXYLTDPBqQxJZfFSMHAIUEO\x1b\x11\x18\x19";
char wylie_decoded[34][7] = {
                        "kh",
                        "gh",
                        "ng",
                        "ch",
                        "jh",
                        "ny",
                        "T",
                        "Th",
                        "D",
                        "Dh",
                        "N",
                        "th",
                        "dh",
                        "ph",
                        "bh",
                        "ts",
                        "tsh",
                        "dz",
                        "dzh",
                        "zh",
                        "S",
                        "kS",
                        "sh",
                        "M",
                        "H",
                        "A",
                        "I",
                        "U",
                        "ai",
                        "au",
                        "r-i",
                        "r-I",
                        "l-i",
                        "l-I"
                     } ;
char ostr  [500] ;
/*uni3_trans*/
void prterr(char * errmsg)
{
int i ;
  fprintf(stderr, errmsg) ;
       for (i = 0 ; i < strlen(errmsg); i = i + 1)
          {
             strncat(ostr, &(errmsg[i]), 1) ;
             strcat(ostr, "\xff") ;
          }
}
char * towylie(char * syll_ptr)
{
  char * wylie_encoded_ptr ;
  strcpy(wylie_syll, "" ) ;
  while ( *syll_ptr != '\0')
    {
      wylie_encoded_ptr = strchr(wylie_encoded, syll_ptr[0]) ;
      if ( wylie_encoded_ptr == NULL )
        {
          strncat( wylie_syll, syll_ptr, 1) ;
        }
      else
        {
          strcat( wylie_syll, wylie_decoded[wylie_encoded_ptr -
                                            wylie_encoded] ) ;
        }
      syll_ptr = syll_ptr + 1 ;
    }
    return (wylie_syll) ;
}
void parse_syll (char * syll)
{
  int vowel_pos ;
  int j ;
  int front_vowel ;
  int vowel_count ;
  int stacking_mode ;
  char * first_vowel_ptr ;
  char * last_vowel_ptr ;
  char * temp_vowel_ptr ;
  char * digit_ptr ;
  char input_prefix_cluster[5] ;
  char output_prefix_cluster[5] ;
/*                                                                */
/*Initialise output slots*/
  for ( j = 0 ; j < sizeof(slot); j = j + 1)
    {
      slot[j] = ' ' ;
    }
  for ( j = 0 ; j < sizeof(vowelslot); j = j + 1)
    {
      vowelslot[j] = ' ' ;
    }
/*Check for digit and process it                     */
  digit_ptr = strpbrk(syll, digits) ;
  while ( (digit_ptr != NULL)
        && ( digit_ptr == syll ) )
    {
      strncat(ostr, &digit_ptr[0], 1) ;
      strcat(ostr, "\xff") ;
      syll = syll + 1 ;
      digit_ptr = strpbrk(syll, digits) ;
    }
  if ( *syll == '\0') return ;
/*Find first vowel and process it                     */
  first_vowel_ptr = strpbrk(syll, vowels) ;
  if ( first_vowel_ptr == NULL )
    {
      prterr("\n") ;
      prterr(towylie(syll)) ;
      prterr(" - Illegal tibetan syllable - no vowel\n") ;
      return ;
    }
  vowel_count = 1 ;
  vowelslot[1] = *first_vowel_ptr ;
  last_vowel_ptr = first_vowel_ptr ;
  temp_vowel_ptr = strpbrk(last_vowel_ptr+1, vowels) ;
  j = 2;
/* now count the vowels */
  while ( temp_vowel_ptr != NULL )
    {
      last_vowel_ptr = temp_vowel_ptr ;
      vowelslot[j] = *temp_vowel_ptr;
      temp_vowel_ptr = strpbrk(last_vowel_ptr+1, vowels) ;
      j = j + 1 ;
      vowel_count = vowel_count + 1 ;
    }
  if (vowel_count == 1)
    {
      front_vowels = &(normal_mode_front_vowels[0]) ;
      vowels = &(normal_mode_vowels[0]) ;
      suffix = &(normal_mode_suffix[0]) ;
      postsuffix = &(normal_mode_postsuffix[0]) ;
    }
  else
    {
      front_vowels = &(multi_vowel_mode_front_vowels[0]) ;
      vowels = &(multi_vowel_mode_vowels[0]) ;
      suffix = &(multi_vowel_mode_suffix[0]) ;
      postsuffix = &(multi_vowel_mode_postsuffix[0]) ;
    }
/*now process the syllable according to number of vowels*/
  vowel_pos = 1 ;
  stacking_mode = 0 ;
  ostr[0] = '\0' ;
  while ( vowel_pos <= vowel_count)
    {
      strcpy(input_prefix_cluster,"") ;
      strcpy(output_prefix_cluster,"     ") ;
/*check if suffix letter is valid and process suffix              */
  if (
          ( strlen(syll) > (1 + first_vowel_ptr - syll) ) &&
          (first_vowel_ptr == last_vowel_ptr)  /*not in multivowel words */
     )
    {
      if (
          strchr(suffix,*(last_vowel_ptr + 1)) != NULL
         )
        {
          slot[6] = *(last_vowel_ptr + 1 ) ;
        }
      else
        {
          prterr("\n") ;
          prterr(towylie(syll)) ;
          prterr(" - Illegal tibetan syllable - bad suffix\n")          ;
          return ;
        }
    }
/*check if postsuffix letter is valid and process it*/
  if (
          ( strlen(syll) > (2 + first_vowel_ptr - syll) ) &&
          ( vowel_count == 1 )/* post suffix not in multivowel words */
     )
    {
      if (
          strchr(postsuffix,*(last_vowel_ptr + 2)) != NULL
         )
        {
          slot[7] = *(last_vowel_ptr + 2 ) ;
        }
      else
        {
              prterr("\n") ;
              prterr(towylie(syll)) ;
              prterr(" - Illegal tibetan syllable - bad postsuffix\n") ;
              return ;
        }
    }
/* Now check for a front-vowel                     */
      front_vowel = 0 ;
      if ( first_vowel_ptr == syll )
        {
          front_vowel = 1 ;
          switch (*first_vowel_ptr)
            {
              case '\x1b':
              case '\x11':
                {
                  slot[3] = 'a' ;
                  break ;
                }
              case '\x18':
              case '\x19':
                {
                  slot[3] = 'a' ;
                  break ;
                }
              default:
                {
                  slot[3] = 'a' ;
                  break ;
                }
            }
        }
/*if any char before the vowel process prefix cluster*/
  if (
      front_vowel == 0
     )
    {
      strncat(input_prefix_cluster, syll, first_vowel_ptr-syll) ;
      for ( j = 0; j<NUMPREFIX; j = j + 1)
        {
          if (
              strcmp(input_prefix_cluster_table[j],
                      input_prefix_cluster)
              == 0
             )
            {
              strcpy(output_prefix_cluster,
                     output_prefix_cluster_table[j]) ;
              break ;
            }
        }
      if (
           j == NUMPREFIX
/*         strcmp(output_prefix_cluster, "    ") == 0    */
         )
        {
          prterr("\n") ;
          prterr(towylie(syll)) ;
          prterr(" - Illegal tibetan syllable - bad prefix cluster\n") ;
          return ;
        }
      slot[1]=output_prefix_cluster[0] ;
      slot[2]=output_prefix_cluster[1] ;
      slot[3]=output_prefix_cluster[2] ;
      slot[4]=output_prefix_cluster[3] ;
      slot[5]=output_prefix_cluster[4] ;
    }
/*leftbrace_count = 0 ;                                           */
/*                                                                */
/*end of parsing, now start output processing                     */
/*output prefix if any                                            */
  if ( slot[1] != ' ' )
    {
          strcat(ostr,
            upper_root_out[strchr(root_in, slot[1]) - root_in]) ;
    }
/*check for superfixed after vowel                   */
  if ( slot[2] != ' ' )
    {
      switch (slot[2])
        {
          case 'r':
            {
            if (stacking_mode == 0)
              {
                strcat(ostr, upper_root_r) ;
              }
            else
              {
                strcat(ostr, lower_root_r) ;
              }
              break ;
            }
          case 'l':
            {
              strcat(ostr, upper_root_l) ;
              break ;
            }
          case 's':
            {
              strcat(ostr, upper_root_s) ;
              break ;
            }
        }
    }
/*no superfixed so do root in Unicode 'upper' chars if not stacking */
if ( (slot[2] == ' ') &&
     (stacking_mode == 0) )
    {
      if ( slot[3] != ' ' )
        {
          strcat(ostr,
            upper_root_out[strchr(root_in, slot[3]) - root_in]) ;
        }
     }
/*superfixed present or stacking,  do root in Unicode 'lower' chars */
else
    {
      if ( slot[3] != ' ' )
        {
          strcat(ostr,
            lower_root_out[strchr(root_in, slot[3]) - root_in]) ;
        }
    }
/*check for subfixed after outputting root                       */
  if ( slot[4] != ' ' )
    {
      switch (slot[4])
        {
          case 'y':
            {
#define sub_yata  "\xb1\x0f"
              strcat(ostr,sub_yata) ;
              break ;
            }
          case 'r':
            {
#define sub_rata  "\xb2\x0f"
              strcat(ostr,sub_rata) ;
              break ;
            }
          case 'l':
            {
#define sub_lata  "\xb3\x0f"
              strcat(ostr,sub_lata) ;
              break ;
            }
          case 'w':
            {
#define sub_wazur "\xad\x0f"
              strcat(ostr,sub_wazur) ;
              break ;
            }
        }
      slot[4] = ' ';
    }
/*check for subsubfixed after outputting root                       */
  if ( slot[5] != ' ' )
    {
      switch (slot[5])
        {
          case 'w':
            {
#define sub_wazur "\xad\x0f"
              strcat(ostr,sub_wazur) ;
              break ;
            }
        }
      slot[5] = ' ';
    }
/*do vowels                                       */
      stacking_mode = 0 ;
      switch (vowelslot[vowel_pos])
        {
          case 'a':
            {
              break ;
            }
          case 'A':
            {
#define vowel_A "\x71\x0f"
              strcat(ostr, vowel_A) ;
              break ;
            }
          case 'i':
            {
#define vowel_i "\x72\x0f"
              strcat(ostr, vowel_i) ;
              break ;
            }
          case 'I':
            {
#define vowel_I "\x71\x0f\x72\x0f"
              strcat(ostr, vowel_I) ;
              break ;
            }
          case 'u':
            {
#define vowel_u "\x74\x0f"
              strcat(ostr, vowel_u) ;
              break ;
            }
          case 'U':
            {
#define vowel_U "\x71\x0f\x74\x0f"
              strcat(ostr, vowel_U) ;
              break ;
            }
          case 'e':
            {
#define vowel_e "\x7a\x0f"
              strcat(ostr, vowel_e) ;
              break ;
            }
          case 'E':
            {
#define vowel_ai "\x7b\x0f"
              strcat(ostr, vowel_ai) ;
              break ;
            }
          case 'o':
            {
#define vowel_o "\x7c\x0f"
              strcat(ostr, vowel_o) ;
              break ;
            }
          case 'O':
            {
#define vowel_au "\x7d\x0f"
              strcat(ostr, vowel_au) ;
              break ;
            }
          case '\x1b':
            {
#define vowel_ri "\xb2\x0f\x80\x0f"
              strcat(ostr, vowel_ri) ;
              break ;
            }
          case '\x11':
            {
#define vowel_rI "\xb2\x0f\x71\x0f\x80\x0f"
              strcat(ostr, vowel_rI) ;
              break ;
            }
          case '\x18':
            {
#define vowel_li "\xb3\x0f\x80\x0f"
              strcat(ostr, vowel_li) ;
              break ;
            }
          case '\x19':
            {
#define vowel_lI "\xb3\x0f\x71\x0f\x80\x0f"
              strcat(ostr, vowel_lI) ;
              break ;
            }
          case '\'':    /*achung as a semivowel*/
            {
              strcat(ostr, upper_root_achung) ;
              break ;
            }
          case '+':  /*use + as a non-breaking space in bracket mode*/
            {
              stacking_mode = 1 ;
              break ;
            }
        }
/*do suffix                                   */
  if ( slot[6] != ' ' )
    {
          strcat(ostr,
            upper_root_out[strchr(root_in, slot[6]) - root_in]) ;
    }
/*do postsuffix                              */
  if ( slot[7] != ' ' )
    {
      strcat(ostr,
        upper_root_out[strchr(root_in, slot[7]) - root_in]) ;
    }
/*now process closing if multiple vowels present              */
        if (vowel_count > 1)
          {
            syll = strpbrk(syll, vowels) + 1;
            first_vowel_ptr = strpbrk(syll, vowels) ;
          }
/*Initialise output slots*/
  for ( j = 0 ; j < sizeof(slot); j = j + 1)
    {
      slot[j] = ' ' ;
    }
  vowel_pos = vowel_pos + 1 ;
    }
}

int uni3_trans (char *iptr, char * ostr)
{
   int i ;
   int syll_len ;
   char * tok_end_ptr ;
/* return first charcter of input string as a string of length one */
/* function value is length of input 'syllable' translated  */
   *ostr = '\0' ;
   *syll = '\0' ;
   if  (strncmp(iptr, "{", 1) == 0)
     {
       prterr("\n") ;
       prterr("Illegal use of { \n") ;                       ;
       return 1 ;
     }
   if (strncmp(iptr, "}", 1) == 0)
     {
       prterr("\n") ;
       prterr("Illegal use of } \n") ;                       ;
       return 1 ;
     }
/* if ((strpbrk(iptr, "\n /|") - iptr) == 0)    */
   if ((strpbrk(iptr, "\n /,") - iptr) == 0)
     {
       if (
           (strcmpl(iptr, "/") == 0)
          )
         {
#define punc_tsheg  "\x0b\x0f"
            strcat(ostr, punc_tsheg) ;
            return 1 ;
         }
       if (
           (strcmpl(iptr, ",") == 0)
          )
         {
#define punc_shad  "\x0d\x0f"
            strcat(ostr, punc_shad) ;
            return 1 ;
         }
       if (
           (strcmpl(iptr, "\n") == 0) &&
           (strcmpl(iptr-1, "/") != 0) &&
/*         (strcmpl(iptr-1, "|") != 0) &&       */
           (strcmpl(iptr-1, ",") != 0) &&
           (strcmpl(iptr-1, "}") != 0)
          )
         {
            strcat(ostr, punc_tsheg) ;
#define uni_newline "\x0d\xff\x0a\xff"
            strcat(ostr, uni_newline) ;
            return 1;
         }
       if (
           (strcmpl(iptr, "\n") == 0)
          )
         {
            strcat(ostr, uni_newline) ;
            return 1;
         }
       strncat(ostr, iptr, 1) ;     /* what's this for? */
       return (1) ;
     }
/* tok_end_ptr = strpbrk(iptr,"\n /|{}") ;      */
   tok_end_ptr = strpbrk(iptr,"\n /,{}()") ;
   if ( tok_end_ptr == NULL)
     {
       if ( strlen(iptr) > (sizeof(syll)-1) )
         {
           prterr("\n") ;
           prterr(" - Syllable too long - truncated\n") ;
           prterr("\n") ;
         }
       strncat(syll, iptr, sizeof(syll)-1 ) ;
       parse_syll(syll) ;
     }
   if (*tok_end_ptr == '(')
     {
       parse_syll("(") ;
       return (1) ;
     }
   if (*tok_end_ptr == ')')
     {
       if ( (tok_end_ptr - iptr) > (sizeof(syll)-1) )
         {
           prterr("\n") ;
           prterr(" - Syllable too long - truncated\n") ;
           prterr("\n") ;
           strncat(syll, iptr, (sizeof(syll)-1) ) ;
         }
       else
           strncat(syll, iptr, (tok_end_ptr - iptr) ) ;
       parse_syll(syll) ;
       parse_syll(")") ;
       syll_len = strlen(syll) ;
       return (syll_len+1) ;
     }
   if (*tok_end_ptr == '+')
     {
       strncat(syll, iptr, (tok_end_ptr - iptr+1) ) ;
       parse_syll(syll) ;
     }
   else
     {
       strncat(syll, iptr, (tok_end_ptr - iptr) ) ;
       parse_syll(syll) ;
     }
/* strcat( ostr, syll) ;   */
   syll_len = strlen(syll) ;
   return (syll_len) ;
}



/*filecopy*/
void filecopy (FILE *ifp, FILE *ofp)
{
   char *iptr ;
   char *optr ;
   int i ;

/*write the unicode file marker */
   fputs("\xff\xfe", ofp) ;
   iptr = fgets(iline, 4095, ifp) ;
   while (iptr != NULL)
      {
       *oline = '\0' ;
       optr = oline ;
       /* loop until end-of-line*/
       if ( *iptr == '\n' )
         {
/* do a Unicode Newline        */
           fputc('\x0d', ofp) ;
           fputc('\x00', ofp) ;
           fputc('\x0a', ofp) ;
           fputc('\x00', ofp) ;
/*         strcat( optr, iptr) ;  */
         }
       else
         {
           while (
                  *iptr != '\0'
                 )
             {
               iptr = iptr + uni3_trans( iptr , ostr) ;
               strcat( optr, ostr) ;
             }
         }

       for (i = 0 ; i < strlen(oline); i = i + 1)
          {
             if ( oline[i] == '\xff')
                {
/* replace x'ff' with x'00' to generate 'real' unicode */
                   fputc('\x00', ofp) ;
                }
             else
                fputc(oline[i], ofp) ;
          }
       iptr = fgets(iline, 4095, ifp) ;

      }

}


main ( int argc, char*argv[] )
{
   void filecopy(FILE *, FILE *) ;

   if (argc != 3) /* wrong args */
      {
       printf("unitran3: wrong number of arguments");
       exit(1) ;
      }
   else
      {
       if ( strcmpl( *(argv + 1), "stdin:") == 0 )
         {
           ifp = stdin ;
         }
       else
         {
           if ((ifp = fopen(*(argv + 1), "r") ) == NULL)
             {
              printf("unitran3 can't open %s\n", *(argv + 1)) ;
              exit(1) ;
             }
         }
       if ( strcmpl( *(argv + 2), "stdout:") == 0 )
         {
           ofp = stdout ;
         }
       else
         {
           if ((ofp = fopen(*(argv + 2), "wb" )) == NULL)
             {
              printf("unitran3 can't open %s\n", *(argv + 2)) ;
              exit(1) ;
             }
         }

/*     iptr = fgets(iline, 4095, ifp) ;   /**/
       filecopy( ifp, ofp ) ;
       fclose (ifp);
       fclose (ofp);
      }
   exit(0);
}
